#include "Field.h"
#include "Piece.h"
#include "pieces/Pawn.h"
#include "pieces/King.h"
#include "pieces/Queen.h"
#include "pieces/Rook.h"
#include "pieces/Bishop.h"
#include "pieces/Knight.h"

CField::CField() {
  for(int y = 0; y < 8; y++) {
    for(int x = 0; x < 8; x++) {
      fields[x][y] = nullptr;
    }
  }

  reset();
}

CField::~CField() {
  // reset previous pieces
  for(int y = 0; y < 8; y++) {
    for(int x = 0; x < 8; x++) {
      if(fields[x][y]) {
        delete fields[x][y];
      }
    }
  }
}

void CField::reset() {

  // reset previous pieces
  for(int y = 0; y < 8; y++) {
    for(int x = 0; x < 8; x++) {
      if(fields[x][y]) {
        delete fields[x][y];
      }
      fields[x][y] = 0;
    }
  }


  // White
  fields[0][0] = new CRook(0, 0, Color::WHITE);
  fields[7][0] = new CRook(7, 0, Color::WHITE);

  fields[1][0] = new CKnight(1, 0, Color::WHITE);
  fields[6][0] = new CKnight(6, 0, Color::WHITE);

  fields[2][0] = new CBishop(2, 0, Color::WHITE);
  fields[5][0] = new CBishop(5, 0, Color::WHITE);

  fields[3][0] = new CQueen(3, 0, Color::WHITE);
  fields[4][0] = (m_pKingWhite = new CKing(4, 0, Color::WHITE));
  for(unsigned int x = 0; x < 8; x++)
    fields[x][1] = new CPawn(x, 1, Color::WHITE);

  // Empty Space
  for(unsigned x = 0; x < 8; x++)
    for(unsigned y = 2; y < 6; y++)
      fields[x][y] = nullptr;

  // Black
  fields[0][7] = new CRook(0, 7, Color::BLACK);
  fields[7][7] = new CRook(7, 7, Color::BLACK);

  fields[1][7] = new CKnight(1, 7, Color::BLACK);
  fields[6][7] = new CKnight(6, 7, Color::BLACK);

  fields[2][7] = new CBishop(2, 7, Color::BLACK);
  fields[5][7] = new CBishop(5, 7, Color::BLACK);

  fields[3][7] = new CQueen(3, 7, Color::BLACK);
  fields[4][7] = (m_pKingBlack = new CKing(4, 7, Color::BLACK));
  for(unsigned int x = 0; x < 8; x++)
    fields[x][6] = new CPawn(x, 6, Color::BLACK);
}

ostream& operator<<(ostream &os, const CField &field) {
  os << "  ";
  for(int x = 0; x < 8; x++)
    os << "  " << (char)((int)'A' + x);
  os << endl;

  for(int y = 7; y >= 0; y--) {
    os << " " << y + 1;
    for(int x = 0; x < 8; x++) {
      os << "  ";
      if(field.fields[x][y] != nullptr)
        os << *field.fields[x][y];
      else
        os << " ";
    }
    os << endl;
  }

  return os;
}

bool CField::checkBounds(int x, int y) {
  return (x >= 0 && y >= 0 && x < 8 && y < 8);
}

bool CField::isAttacked(int tarX, int tarY, int color) {
  if(!checkBounds(tarX, tarY))
    return false;

  for(int x = 0; x < 8; x++) {
    for(int y = 0; y < 8; y++) {

      if((x != tarX || y != tarY) &&
         fields[x][y] != nullptr &&
         fields[x][y]->getColor() == color &&
         fields[x][y]->canMove(this, tarX, tarY))
         return true;
    }
  }

  return false;
}

bool CField::move(int fromX, int fromY, int toX, int toY) {
  if(!checkBounds(fromX, fromY) || !checkBounds(fromX, fromY))
    return false;

  if(fromX == toX && fromY == toY) {
    cout << "Source and Target positions are the same" << endl;
    return false;
  }

  CPiece *movingPiece = getPiece(fromX, fromY);
  if(movingPiece == nullptr)
    return false;

  CPiece *targetPiece = getPiece(toX, toY);
  if(targetPiece != nullptr) {
    if(targetPiece->getType() == CPiece::Type::KING) {
      cout << "You cannot attack the king" << endl;
      return false;
    }
    else if(targetPiece->getColor() == movingPiece->getColor()) {
      cout << "You cannot attack your own pieces" << endl;
      return false;
    }
  }

  if(!movingPiece->moveTo(this, toX, toY)) {
    cout << "This piece cannot move there." << endl;
    return false;
  }

  int activeColor = movingPiece->getColor();
  CKing *pOwnKing = (activeColor == Color::WHITE ? m_pKingWhite : m_pKingBlack);
  if(isAttacked(pOwnKing->m_X, pOwnKing->m_Y, 1 - activeColor)) {
    // we cannot move as our king is attacked, reset this move!
    cout << "You king is still attacked." << endl;
    setPiece(fromX, fromY, movingPiece);
    setPiece(toX, toY, targetPiece);
    return false;
  }

  movingPiece->m_HasMoved = true;
  if(targetPiece != nullptr)
    delete targetPiece;

  return true;
}

void CField::setPiece(int x, int y, CPiece *pPiece) {
  if(checkBounds(x, y)) {
    fields[x][y] = pPiece;
    if(pPiece) {
      pPiece->m_X = x;
      pPiece->m_Y = y;
    }
  }
}

int CField::normalalize(int i) {
  return (i == 0 ? 0 : i / abs(i));
}

/*
* returns false, if no intersection occurs, true otherwise
*/
bool CField::intersect(int fromX, int fromY, int toX, int toY, bool includeTarget) {
  if(!checkBounds(fromX, fromY) || !checkBounds(fromX, fromY))
    return false;

  int dirX = toX - fromX;
  int dirY = toY - fromY;

  // No move
  if(dirX == 0 && dirY == 0)
    return false;

  // We cannot calculate intersection for this direction, e.g. dirX = 5, dirY = 3
  if(dirX != 0 && dirY != 0 && abs(dirX) != abs(dirY))
    return false;

  dirX = normalalize(dirX);
  dirY = normalalize(dirY);

  int x = fromX + dirX;
  int y = fromY + dirY;

  for(; !(x == toX && y == toY); x += dirX, y += dirY) {
    if(fields[x][y] != nullptr)
      return true;
  }

  return (includeTarget ? fields[toX][toY] != nullptr : false);
}
